/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

/* this is a subset of mpack for SDPA-GMP only */
/* http://accc.riken.jp/maho/mpack/index.html */

#include <gmpxx.h>

using namespace std;

#if !defined (__MPACK_ERRNO__)
extern int mpack_errno;
#endif

#if !defined (MPACK_DEFAULT_PRECISION)
#define MPACK_DEFAULT_PRECISION 256
#endif

//implementation of sign transfer function.
inline mpf_class Msign(mpf_class a, mpf_class b)
{
  mpf_class mtmp;
  mpf_abs(mtmp.get_mpf_t(),a.get_mpf_t());
  if (b!=0.0) {
    mtmp=mpf_sgn(b.get_mpf_t())*mtmp;
  }
  return mtmp;
}

/* Initalization */
void mpack_initialize(int precision);

/* old routine */
void Rset(int N, mpf_class a, mpf_class * x, int step);

/* LEVEL 1 MBLAS */
mpf_class Rdot(int n, mpf_class * dx, int incx, mpf_class * dy, int incy);
void Rcopy(int n, mpf_class * dx, int incx, mpf_class * dy, int incy);
void Raxpy(int n, mpf_class da, mpf_class * dx, int incx, mpf_class * dy,
    int incy);
void Rscal(int n, mpf_class ca, mpf_class * cx, int incx);
mpf_class Rnrm2(int n, mpf_class * x, int incx);
int Mlsame(const char *a, const char *b);
void Mxerbla(const char *srname, int info);
void Rswap(int n, mpf_class * dx, int incx, mpf_class * dy, int incy);

/* LEVEL 2 MBLAS */
void Rtrmv(const char *uplo, const char *trans, const char *diag, int n,
    mpf_class * A, int lda, mpf_class * x, int incx);
void Rtrsv(const char *uplo, const char *trans, const char *diag, int n,
	   mpf_class * A, int lda, mpf_class * x, int incx);
void Rgemv(const char *trans, int m, int n, mpf_class alpha, mpf_class * A,
	   int lda, mpf_class * x, int incx, mpf_class beta, mpf_class * y, int incy);
void Rsymv(const char *uplo, int n, mpf_class alpha, mpf_class * A, int lda,
	   mpf_class * x, int incx, mpf_class beta, mpf_class * y, int incy);
void Rsyr2(const char *uplo, int n, mpf_class alpha, mpf_class * x, int incx,
	   mpf_class * y, int incy, mpf_class * A, int lda);
void Rger(int m, int n, mpf_class alpha, mpf_class * x, int incx,
	 mpf_class * y, int incy, mpf_class * A, int lda);

/* LEVEL 3 MBLAS */
void Rtrmm(const char *side, const char *uplo, const char *transa,
    const char *diag, int m, int n, mpf_class alpha, mpf_class * A, int lda,
    mpf_class * B, int ldb);
void Rtrsm(const char *side, const char *uplo, const char *transa,
    const char *diag, int m, int n, mpf_class alpha, mpf_class * A, int lda,
    mpf_class * B, int ldb);
void Rgemm(const char *transa, const char *transb, int m, int n, int k,
	   mpf_class alpha, mpf_class * A, int lda, mpf_class * B, int ldb,
	   mpf_class beta, mpf_class * C, int ldc); 
void Rsyr2k(const char *uplo, const char *trans, int n, int k, mpf_class alpha,
	    mpf_class * A, int lda, mpf_class * B, int ldb, mpf_class beta,
	    mpf_class * C, int ldc);
void Rsyrk(const char *uplo, const char *trans, int n, int k, mpf_class alpha,
	   mpf_class * A, int lda, mpf_class beta, mpf_class * C, int ldc);

