/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

/* mlapack prototypes */

/* this is a subset of mpack for SDPA-GMP only */
/* http://accc.riken.jp/maho/mpack/index.html */

#include <math.h> //for log10

inline mpf_class mpf_approx_log10(mpf_class x)
{
  double d;
  double ln10_app;
  signed long int exp;
  d=mpf_get_d_2exp(&exp, x.get_mpf_t());
  ln10_app = (double) exp + log10(d);
  return ln10_app; 
}

void Rsteqr(const char *compz, int n, mpf_class * d, mpf_class * e,
	    mpf_class * Z, int ldz, mpf_class * work, int *info);
void Rsyev(const char *jobz, const char *uplo, int n, mpf_class * A, 
            int lda, mpf_class * w, mpf_class * work, int *lwork, int *info);
void Rpotrf(const char *uplo, int n, mpf_class * A, int lda, int *info);
int iMlaenv(int ispec, const char *name, const char *opts, int n1, int n2,
	    int n3, int n4);
mpf_class Rlamch(const char *cmach);
mpf_class Rlansy(const char *norm, const char *uplo, int n, mpf_class * A,
		 int lda, mpf_class * work);
void Rlascl(const char *type, int kl, int ku, mpf_class cfrom, mpf_class cto,
	    int m, int n, mpf_class * A, int lda, int *info);
void Rsytrd(const char *uplo, int n, mpf_class * A, int lda, mpf_class * d,
	    mpf_class * e, mpf_class * tau, mpf_class * work, int lwork, int *info);
void Rsterf(int n, mpf_class * d, mpf_class * e, int *info);
void Rorgtr(const char *uplo, int n, mpf_class * a, int lda, mpf_class * tau,
	    mpf_class * work, int lwork, int *info);
void Rlatrd(const char *uplo, int n, int nb, mpf_class * A, int lda,
	    mpf_class * e, mpf_class * tau, mpf_class * w, int ldw);
void Rsytd2(const char *uplo, int n, mpf_class * A, int lda, mpf_class * d,
	    mpf_class * e, mpf_class * tau, int *info);
mpf_class Rlanst(const char *norm, int n, mpf_class * d, mpf_class * e);
void Rlae2(mpf_class a, mpf_class b, mpf_class c, mpf_class * rt1,
	   mpf_class * rt2);
mpf_class Rlapy2(mpf_class x, mpf_class y);
void Rlasrt(const char *id, int n, mpf_class * d, int *info);
void Rorgql(int m, int n, int k, mpf_class * A, int lda, mpf_class * tau,
	    mpf_class * work, int lwork, int *info);
void Rorgqr(int m, int n, int k, mpf_class * A, int lda, mpf_class * tau,
	    mpf_class * work, int lwork, int *info);
void Rlarfg(int N, mpf_class * alpha, mpf_class * x, int incx,
	    mpf_class * tau);
void Rlassq(int n, mpf_class * x, int incx, mpf_class * scale,
	    mpf_class * sumsq);
void Rorg2l(int m, int n, int k, mpf_class * A, int lda, mpf_class * tau,
	    mpf_class * work, int *info);
void Rlarft(const char *direct, const char *storev, int n, int k,
	    mpf_class * v, int ldv, mpf_class * tau, mpf_class * t, int ldt);
void Rlarfb(const char *side, const char *trans, const char *direct,
	    const char *storev, int m, int n, int k, mpf_class * V, int ldv,
	    mpf_class * T, int ldt, mpf_class * C, int ldc, mpf_class * work,
	    int ldwork);
void Rorg2r(int m, int n, int k, mpf_class * A, int lda, mpf_class * tau,
	    mpf_class * work, int *info);
void Rlarf(const char *side, int m, int n, mpf_class * v, int incv,
	   mpf_class tau, mpf_class * C, int ldc, mpf_class * work);
void Rpotf2(const char *uplo, int n, mpf_class * A, int lda, int *info);
void Rlaset(const char *uplo, int m, int n, mpf_class alpha, mpf_class beta,
	    mpf_class * A, int lda);
void Rlaev2(mpf_class a, mpf_class b, mpf_class c, mpf_class * rt1,
	    mpf_class * rt2, mpf_class * cs1, mpf_class * sn1);
void Rlasr(const char *side, const char *pivot, const char *direct, int m,
	   int n, mpf_class * c, mpf_class * s, mpf_class * A, int lda);
void Rlartg(mpf_class f, mpf_class g, mpf_class * cs, mpf_class * sn,
	    mpf_class * r);
